# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the object code for the DrawingView object (OBSOLETE).

This module is obsolete, since the Drawing Workbench stopped
being developed in v0.17.
The TechDraw Workbench replaces Drawing, and it no longer requires
a `DrawingView` object to display objects in a drawing sheet.

This module is still provided in order to be able to open older files
that use this `DrawingView` object. However, a GUI tool to create
this object should no longer be available.
"""
## @package drawingview
# \ingroup draftobjects
# \brief Provides the object code for the DrawingView object (OBSOLETE).

## \addtogroup draftobjects
# @{
from PySide.QtCore import QT_TRANSLATE_NOOP

import draftfunctions.svg as get_svg
import draftfunctions.dxf as get_dxf
import draftutils.utils as utils
import draftutils.groups as groups

from draftobjects.base import DraftObject


class DrawingView(DraftObject):
    """The DrawingView object. This class is OBSOLETE.

    This object was used with the Drawing Workbench, but since this workbench
    because obsolete in v0.17, the object should no longer be used.
    It is retained for compatibility purposes, that is, to open older
    files that may contain this object.

    To produce 2D drawings, use TechDraw Workbench.
    """

    def __init__(self, obj):
        super(DrawingView, self).__init__(obj, "DrawingView")

        _tip = QT_TRANSLATE_NOOP("App::Property",
                                 "The linked object")
        obj.addProperty("App::PropertyLink",
                        "Source",
                        "Base",
                        _tip)

        _tip = QT_TRANSLATE_NOOP("App::Property",
                                 "Projection direction")
        obj.addProperty("App::PropertyVector",
                        "Direction",
                        "Shape View",
                        _tip)

        _tip = QT_TRANSLATE_NOOP("App::Property",
                                 "The width of the lines inside this object")
        obj.addProperty("App::PropertyFloat",
                        "LineWidth",
                        "View Style",
                        _tip)
        obj.LineWidth = 0.35

        _tip = QT_TRANSLATE_NOOP("App::Property",
                                 "The size of the texts inside this object")
        obj.addProperty("App::PropertyLength",
                        "FontSize",
                        "View Style",
                        _tip)
        obj.FontSize = 12

        _tip = QT_TRANSLATE_NOOP("App::Property",
                                 "The spacing between lines of text")
        obj.addProperty("App::PropertyLength",
                        "LineSpacing",
                        "View Style",
                        _tip)

        _tip = QT_TRANSLATE_NOOP("App::Property",
                                 "The color of the projected objects")
        obj.addProperty("App::PropertyColor",
                        "LineColor",
                        "View Style",
                        _tip)

        _tip = QT_TRANSLATE_NOOP("App::Property",
                                 "Shape Fill Style")
        obj.addProperty("App::PropertyEnumeration",
                        "FillStyle",
                        "View Style",
                        _tip)
        obj.FillStyle = ['shape color'] + list(utils.svgpatterns().keys())

        _tip = QT_TRANSLATE_NOOP("App::Property",
                                 "Line Style")
        obj.addProperty("App::PropertyEnumeration",
                        "LineStyle",
                        "View Style",
                        _tip)
        obj.LineStyle = ['Solid', 'Dashed', 'Dotted', 'Dashdot']

        _tip = QT_TRANSLATE_NOOP("App::Property",
                                 "If checked, source objects are displayed "
                                 "regardless of being visible in the 3D model")
        obj.addProperty("App::PropertyBool",
                        "AlwaysOn",
                        "View Style",
                        _tip)

    def execute(self, obj):
        """Execute when the object is created or recomputed."""
        result = ""
        if hasattr(obj, "Source") and obj.Source:
            if hasattr(obj, "LineStyle"):
                ls = obj.LineStyle
            else:
                ls = None
            if hasattr(obj, "LineColor"):
                lc = obj.LineColor
            else:
                lc = None
            if hasattr(obj, "LineSpacing"):
                lp = obj.LineSpacing
            else:
                lp = None

            if obj.Source.isDerivedFrom("App::DocumentObjectGroup"):
                svg = ""
                objs = groups.get_group_contents([obj.Source])
                for o in objs:
                    v = o.ViewObject.isVisible()
                    if hasattr(obj, "AlwaysOn") and obj.AlwaysOn:
                        v = True
                    if v:
                        svg += get_svg.get_svg(o,
                                               obj.Scale,
                                               obj.LineWidth,
                                               obj.FontSize.Value,
                                               obj.FillStyle,
                                               obj.Direction, ls, lc, lp)
            else:
                svg = get_svg.get_svg(obj.Source,
                                      obj.Scale,
                                      obj.LineWidth,
                                      obj.FontSize.Value,
                                      obj.FillStyle,
                                      obj.Direction, ls, lc, lp)

            result += '<g id="' + obj.Name + '"'
            result += ' transform="'
            result += 'rotate(' + str(obj.Rotation) + ','
            result += str(obj.X) + ',' + str(obj.Y)
            result += ') '
            result += 'translate(' + str(obj.X) + ',' + str(obj.Y) + ') '
            result += 'scale(' + str(obj.Scale) + ',' + str(-obj.Scale)
            result += ')'
            result += '">'
            result += svg
            result += '</g>'
        obj.ViewResult = result

    def getDXF(self, obj):
        """Return a DXF fragment."""
        return get_dxf.get_dxf(obj)


# Alias for compatibility with v0.18 and earlier
_DrawingView = DrawingView

## @}



# ***************************************************************************
# *   (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>                  *
# *   (c) 2009, 2010 Ken Cline <cline@frii.com>                             *
# *   (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de>           *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides GUI tools to create parametric Array objects (OBSOLETE).

These commands were replaced by individual commands `Draft_OrthoArray`,
`Draft_PolarArray`, and `Draft_CircularArray` which launch their own
task panel, and provide a more useful way of creating the desired array.
"""
## @package gui_array_simple
# \ingroup draftguitools
# \brief Provides GUI tools to create parametric Array objects (OBSOLETE).

## \addtogroup draftguitools
# @{
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCADGui as Gui
import Draft_rc
import draftguitools.gui_base_original as gui_base_original
import draftguitools.gui_tool_utils as gui_tool_utils
from draftutils.messages import _msg
from draftutils.translate import translate

# The module is used to prevent complaints from code checkers (flake8)
True if Draft_rc.__name__ else False


class Array(gui_base_original.Modifier):
    """Gui Command for the original simple Array tool.

    Parameters
    ----------
    use_link: bool, optional
        It defaults to `False`. If it is `True`, the created object
        will be a `Link array`.
    """

    def __init__(self, use_link=False):
        super(Array, self).__init__()
        self.use_link = use_link

    def GetResources(self):
        """Set icon, menu and tooltip."""

        return {'Pixmap': 'Draft_Array',
                'MenuText': QT_TRANSLATE_NOOP("Draft_Array", "Array"),
                'ToolTip': QT_TRANSLATE_NOOP("Draft_Array", "Creates an array from a selected object.\nBy default, it is a 2x2 orthogonal array.\nOnce the array is created its type can be changed\nto polar or circular, and its properties can be modified.")}

    def Activated(self, name=translate("draft","Array")):
        """Execute when the command is called."""
        super(Array, self).Activated(name=name)
        if not Gui.Selection.getSelection():
            if self.ui:
                self.ui.selectUi()
                _msg(translate("draft", "Select an object to array"))
                self.call = self.view.addEventCallback(
                    "SoEvent",
                     gui_tool_utils.selectObject)
        else:
            self.proceed()

    def proceed(self):
        """Proceed with the command if one object was selected."""
        if Gui.Selection.getSelection():
            obj = Gui.Selection.getSelection()[0]
            Gui.addModule("Draft")
            _cmd = 'Draft.makeArray'
            _cmd += '('
            _cmd += 'FreeCAD.ActiveDocument.' + obj.Name + ', '
            _cmd += 'FreeCAD.Vector(1, 0, 0), '
            _cmd += 'FreeCAD.Vector(0, 1, 0), '
            _cmd += '2, '
            _cmd += '2, '
            _cmd += 'use_link=' + str(self.use_link)
            _cmd += ')'
            _cmd_list = ['obj = ' + _cmd,
                         'Draft.autogroup(obj)',
                         'FreeCAD.ActiveDocument.recompute()']
            self.commit(translate("draft", "Array"),
                        _cmd_list)
        self.finish()


Gui.addCommand('Draft_Array', Array())


class LinkArray(Array):
    """Gui Command for the LinkArray tool based on the simple Array tool."""

    def __init__(self):
        super(LinkArray, self).__init__(use_link=True)

    def GetResources(self):
        """Set icon, menu and tooltip."""

        return {'Pixmap': 'Draft_LinkArray',
                'MenuText': QT_TRANSLATE_NOOP("Draft_LinkArray", "LinkArray"),
                'ToolTip': QT_TRANSLATE_NOOP("Draft_LinkArray", "Like the Array tool, but creates a 'Link array' instead.\nA 'Link array' is more efficient when handling many copies but the 'Fuse' option cannot be used.")}

    def Activated(self):
        """Execute when the command is called."""
        super(LinkArray, self).Activated(name=translate("draft","Link array"))


Gui.addCommand('Draft_LinkArray', LinkArray())

## @}



# ***************************************************************************
# *   (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>                  *
# *   (c) 2009, 2010 Ken Cline <cline@frii.com>                             *
# *   (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de>           *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides GUI tools to enable and disable the working plane grid."""
## @package gui_grid
# \ingroup draftguitools
# \brief Provides GUI tools to enable and disable the working plane grid.

## \addtogroup draftguitools
# @{
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCADGui as Gui
import draftguitools.gui_base as gui_base

from draftutils.translate import translate


class ToggleGrid(gui_base.GuiCommandSimplest):
    """The Draft ToggleGrid command definition.

    If the grid tracker is invisible (hidden), it makes it visible (shown);
    and if it is visible, it hides it.

    It inherits `GuiCommandSimplest` to set up the document
    and other behavior. See this class for more information.
    """

    def __init__(self):
        super(ToggleGrid, self).__init__(name=translate("draft","Toggle grid"))

    def GetResources(self):
        """Set icon, menu and tooltip."""

        d = {'Pixmap': 'Draft_Grid',
             'Accel': "G,R",
             'MenuText': QT_TRANSLATE_NOOP("Draft_ToggleGrid","Toggle grid"),
             'ToolTip': QT_TRANSLATE_NOOP("Draft_ToggleGrid","Toggles the Draft grid on and off."),
             'CmdType': 'ForEdit'}

        return d

    def Activated(self):
        """Execute when the command is called."""
        super(ToggleGrid, self).Activated()

        if hasattr(Gui, "Snapper"):
            Gui.Snapper.setTrackers()
            if Gui.Snapper.grid:
                if Gui.Snapper.grid.Visible:
                    Gui.Snapper.grid.off()
                    Gui.Snapper.forceGridOff = True
                else:
                    Gui.Snapper.grid.on()
                    Gui.Snapper.forceGridOff = False


Gui.addCommand('Draft_ToggleGrid', ToggleGrid())

## @}



# ***************************************************************************
# *   Copyright (c) 2020 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

from __future__ import print_function


__title__ =  "FreeCAD structural IFC export tools"
__author__ = "Yorik van Havre"
__url__ =    "https://www.freecadweb.org"

ALLOW_LINEAR_OBJECTS = True # allow non-solid objects (wires, etc) to become analytic objects?

structural_nodes = {} # this keeps track of nodes during this session
scaling = 1.0 # this keeps track of scaling during this session


def setup(ifcfile,ifcbin,scale):

    """Creates all the needed setup for structural model."""

    global structural_nodes,scaling
    structural_nodes = {}
    scaling = scale
    import ifcopenshell
    uid = ifcopenshell.guid.new
    owh = ifcfile.by_type("IfcOwnerHistory")[0]
    prj = ifcfile.by_type("IfcProject")[0]
    ctx = createStructuralContext(ifcfile)
    if ifcfile.wrapped_data.schema_name == "IFC2X3":
        mod = ifcfile.createIfcStructuralAnalysisModel(uid(),owh,"Structural Analysis Model",None,None,"NOTDEFINED",None,None,None)
    else:
        pla = ifcbin.createIfcLocalPlacement()
        mod = ifcfile.createIfcStructuralAnalysisModel(uid(),owh,"Structural Analysis Model",None,None,"NOTDEFINED",None,None,None,pla)
    rel = ifcfile.createIfcRelDeclares(uid(),owh,None,None,prj,[mod])


def createStructuralContext(ifcfile):

    """Creates an additional geometry context for structural objects. Returns the new context"""

    contexts = ifcfile.by_type("IfcGeometricRepresentationContext")
    # filter out subcontexts
    contexts = [c for c in contexts if c.is_a() == "IfcGeometricRepresentationContext"]
    ctx = contexts[0] # arbitrarily take the first one...
    structcontext = ifcfile.createIfcGeometricRepresentationSubContext('Analysis','Axis',None,None,None,None,ctx,None,"GRAPH_VIEW",None)
    return structcontext


def getStructuralContext(ifcfile):

    """Returns the structural context from the file"""
    for c in ifcfile.by_type("IfcGeometricRepresentationSubContext"):
        if c.ContextIdentifier == "Analysis":
            return c


def createStructuralNode(ifcfile,ifcbin,point):

    """Creates a connection node at the given point"""

    import ifcopenshell
    uid = ifcopenshell.guid.new
    owh = ifcfile.by_type("IfcOwnerHistory")[0]
    ctx = getStructuralContext(ifcfile)
    cpt = ifcbin.createIfcCartesianPoint(tuple(point))
    vtx = ifcfile.createIfcVertexPoint(cpt)
    rep = ifcfile.createIfcTopologyRepresentation(ctx,'Analysis','Vertex',[vtx])
    psh = ifcfile.createIfcProductDefinitionShape(None,None,[rep])
    # boundary conditions serve for ex. to create fixed nodes
    #cnd = ifcfile.createIfcBoundaryNodeCondition("Fixed",ifcfile.createIfcBoolean(True),ifcfile.createIfcBoolean(True),ifcfile.createIfcBoolean(True),ifcfile.createIfcBoolean(True),ifcfile.createIfcBoolean(True),ifcfile.createIfcBoolean(True))
    # for now we don't create any boundary condition
    cnd = None
    pla = ifcbin.createIfcLocalPlacement()
    prd = ifcfile.createIfcStructuralPointConnection(uid(),owh,'Vertex',None,None,pla,psh,cnd,None)
    return prd


def createStructuralMember(ifcfile,ifcbin,obj):

    """Creates a structural member if possible. Returns the member"""

    global structural_nodes
    prd = None
    import Draft
    import Part
    import ifcopenshell
    uid = ifcopenshell.guid.new
    owh = ifcfile.by_type("IfcOwnerHistory")[0]
    ctx = getStructuralContext(ifcfile)
    edges = None
    if Draft.getType(obj) not in ["Structure"]:
        if ALLOW_LINEAR_OBJECTS and obj.isDerivedFrom("Part::Feature"):
            if obj.Shape.Faces:
                return None
            elif not obj.Shape.Edges:
                return None
            else:
                edges = obj.Shape.Edges
    else:
        wire = Part.makePolygon([obj.Placement.multVec(n) for n in obj.Nodes])
        edges = wire.Edges
    if not edges:
        return None
    for edge in edges:
        if len(edge.Vertexes) > 1:
            # we don't care about curved edges just now...
            v0 = edge.Vertexes[0].Point.multiply(scaling)
            v1 = edge.Vertexes[-1].Point.multiply(scaling)
            cp1 = ifcbin.createIfcCartesianPoint(tuple(v0))
            cp2 = ifcbin.createIfcCartesianPoint(tuple(v1))
            upv = ifcbin.createIfcDirection((0,0,1))
            pla = ifcbin.createIfcLocalPlacement()
            vp1 = ifcfile.createIfcVertexPoint(cp1)
            vp2 = ifcfile.createIfcVertexPoint(cp2)
            edg = ifcfile.createIfcEdge(vp1,vp2)
            rep = ifcfile.createIfcTopologyRepresentation(ctx,'Analysis','Edge',[edg])
            psh = ifcfile.createIfcProductDefinitionShape(None,None,[rep])
            prd = ifcfile.createIfcStructuralCurveMember(uid(),owh,obj.Label,None,None,pla,psh,"RIGID_JOINED_MEMBER",upv)
            # check for existing connection nodes
            for v in [v0,v1]:
                vk = tuple(v)
                if vk in structural_nodes:
                    if structural_nodes[vk]:
                        n = structural_nodes[vk]
                    else:
                        # there is another member with same point, create a new node
                        n = createStructuralNode(ifcfile,ifcbin,v)
                        structural_nodes[vk] = n
                    ifcfile.createIfcRelConnectsStructuralMember(uid(),None,None,None,prd,n,None,None,None,None);
                else:
                    # just add the point, no other member using it yet
                    structural_nodes[vk] = None
    return prd


def createStructuralGroup(ifcfile):

    "Assigns all structural objects found in the file to the structural model"""

    import ifcopenshell
    uid = ifcopenshell.guid.new
    owh = ifcfile.by_type("IfcOwnerHistory")[0]
    edges = ifcfile.by_type("IfcStructuralCurveMember")
    verts = ifcfile.by_type("IfcStructuralPointConnection")
    model = ifcfile.by_type("IfcStructuralAnalysisModel")[0]
    if model:
        members = edges + verts
        if members:
            ifcfile.createIfcRelAssignsToGroup(uid(),owh,None,None,members,"PRODUCT",model)


def associates(ifcfile,aobj,sobj):

    """Associates an arch object with a struct object"""

    # This is probably not the right way to do this, ie. relate a structural
    # object with an IfcProduct. Needs to investigate more....

    import ifcopenshell
    uid = ifcopenshell.guid.new
    owh = ifcfile.by_type("IfcOwnerHistory")[0]
    ifcfile.createIfcRelAssignsToProduct(uid(),owh,None,None,[sobj],None,aobj)



from FreeCAD import Vector
import Part

import unittest

class RegressionTests(unittest.TestCase):

    def test_issue_4456(self):
        """
        0004456: Regression : Part.Plane.Intersect do not accept plane as argument
        """
        p1 = Part.Plane()
        p2 = Part.Plane(Vector(0, 0, 0), Vector(1, 0, 0))

        result = p1.intersect(p2)
        line = result.pop()
        self.assertEqual(line.Location, Vector(0, 0, 0))
        self.assertEqual(line.Direction, Vector(0, 1, 0))
        # We should now have empty list...
        with self.assertRaises(IndexError):
            result.pop()



# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Define geometry functions for manipulating shapes in the Draft Workbench.

These functions are used by different object creation functions
of the Draft Workbench, both in `Draft.py` and `DraftTools.py`.
They operate on the internal shapes (`Part::TopoShape`) of different objects
and on their subelements, that is, vertices, edges, and faces.
"""
## \defgroup DRAFTGEOMUTILS DraftGeomUtils
#  \ingroup UTILITIES
#  \brief Shape manipulation utilities for the Draft workbench
#
# Shapes manipulation utilities

## \addtogroup DRAFTGEOMUTILS
#  @{

__title__ = "FreeCAD Draft Workbench - Geometry library"
__author__ = "Yorik van Havre, Jacques-Antoine Gaudin, Ken Cline"
__url__ = ["https://www.freecadweb.org"]

# Doesn't need requisites
from draftgeoutils.linear_algebra import (linearFromPoints,
                                          determinant)

# Needs math, Part, and vector tools
from draftgeoutils.general import PARAMGRP as params
from draftgeoutils.general import NORM

from draftgeoutils.general import (precision,
                                   vec,
                                   edg,
                                   getVerts,
                                   v1,
                                   isNull,
                                   isPtOnEdge,
                                   hasCurves,
                                   isAligned,
                                   getQuad,
                                   areColinear,
                                   hasOnlyWires,
                                   geomType,
                                   isValidPath,
                                   findClosest,
                                   getBoundaryAngles)

# Needs general functions
from draftgeoutils.geometry import (findPerpendicular,
                                    findDistance,
                                    getSplineNormal,
                                    get_spline_normal,
                                    getNormal,
                                    get_normal,
                                    getRotation,
                                    isPlanar,
                                    is_planar,
                                    calculatePlacement,
                                    mirror,
                                    are_coplanar,
                                    is_straight_line)

from draftgeoutils.edges import (findEdge,
                                 orientEdge,
                                 isSameLine,
                                 isLine,
                                 is_line,
                                 invert,
                                 findMidpoint,
                                 getTangent)

from draftgeoutils.faces import (concatenate,
                                 getBoundary,
                                 isCoplanar,
                                 is_coplanar,
                                 bind,
                                 cleanFaces,
                                 removeSplitter)

from draftgeoutils.arcs import (isClockwise,
                                isWideAngle,
                                arcFrom2Pts,
                                arcFromSpline)

from draftgeoutils.cuboids import (isCubic,
                                   getCubicDimensions)

# Needs geometry functions
from draftgeoutils.circle_inversion import (pointInversion,
                                            polarInversion,
                                            circleInversion)

# Needs edges functions
from draftgeoutils.sort_edges import (sortEdges,
                                      sortEdgesOld)

from draftgeoutils.intersections import (findIntersection,
                                         wiresIntersect,
                                         connect,
                                         angleBisection)

from draftgeoutils.wires import (findWires,
                                 findWiresOld,
                                 findWiresOld2,
                                 flattenWire,
                                 superWire,
                                 isReallyClosed,
                                 curvetowire,
                                 curvetosegment,
                                 rebaseWire,
                                 removeInterVertices,
                                 cleanProjection,
                                 tessellateProjection)

# Needs wires functions
from draftgeoutils.fillets import (fillet,
                                   filletWire)

# Needs intersections functions
from draftgeoutils.offsets import (pocket2d,
                                   offset,
                                   offsetWire)

from draftgeoutils.circles import (findClosestCircle,
                                   getCircleFromSpline,
                                   circlefrom1Line2Points,
                                   circlefrom2Lines1Point,
                                   circleFrom2LinesRadius,
                                   circleFrom3LineTangents,
                                   circleFromPointLineRadius,
                                   circleFrom2PointsRadius,
                                   findHomotheticCenterOfCircles,
                                   findRadicalAxis,
                                   findRadicalCenter)

from draftgeoutils.circles_apollonius import (outerSoddyCircle,
                                              innerSoddyCircle,
                                              circleFrom3CircleTangents)

# Needs circles_apollonius functions
# These functions are not imported because they are incomplete;
# they require pre-requisite functions that haven't been written
# from draftgeoutils.circles_incomplete import (circleFrom2tan1pt,
#                                               circleFrom2tan1rad,
#                                               circleFrom1tan2pt,
#                                               circleFrom1tan1pt1rad,
#                                               circleFrom3tan)

##  @}



# ***************************************************************************
# *   (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de>           *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides GUI tools to create Fillet objects between two lines.

TODO: Currently this tool uses the DraftGui widgets. We want to avoid using
this big module because it creates manually the interface.
Instead we should provide its own .ui file and task panel,
similar to the OrthoArray tool.
"""
## @package gui_fillet
# \ingroup draftguitools
# \brief Provides GUI tools to create Fillet objects between two lines.

## \addtogroup draftguitools
# @{
import PySide.QtCore as QtCore
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCADGui as Gui
import Draft
import Draft_rc
import draftutils.utils as utils
import draftguitools.gui_base_original as gui_base_original
import draftguitools.gui_tool_utils as gui_tool_utils

from draftutils.messages import _msg, _err
from draftutils.translate import translate

# The module is used to prevent complaints from code checkers (flake8)
True if Draft_rc.__name__ else False


class Fillet(gui_base_original.Creator):
    """Gui command for the Fillet tool."""

    def __init__(self):
        super(Fillet, self).__init__()
        self.featureName = "Fillet"

    def GetResources(self):
        """Set icon, menu and tooltip.""" 

        return {'Pixmap': 'Draft_Fillet',
                'MenuText': QT_TRANSLATE_NOOP("Draft", "Fillet"),
                'ToolTip': QT_TRANSLATE_NOOP("Draft", "Creates a fillet between two selected wires or edges.")}

    def Activated(self, name=translate("Draft", "Fillet")):
        """Execute when the command is called."""
        super(Fillet, self).Activated(name=name)

        if self.ui:
            self.rad = 100
            self.chamfer = False
            self.delete = False
            label = translate("draft", "Fillet radius")
            tooltip = translate("draft", "Radius of fillet")

            # Call the task panel defined in DraftGui to enter a radius.
            self.ui.taskUi(title=name, icon="Draft_Fillet")
            self.ui.radiusUi()
            self.ui.sourceCmd = self
            self.ui.labelRadius.setText(label)
            self.ui.radiusValue.setToolTip(tooltip)
            self.ui.setRadiusValue(self.rad, "Length")
            self.ui.check_delete = self.ui._checkbox("isdelete",
                                                     self.ui.layout,
                                                     checked=self.delete)
            self.ui.check_delete.setText(translate("Draft",
                                                   "Delete original objects"))
            self.ui.check_delete.show()
            self.ui.check_chamfer = self.ui._checkbox("ischamfer",
                                                      self.ui.layout,
                                                      checked=self.chamfer)
            self.ui.check_chamfer.setText(translate("Draft",
                                                    "Create chamfer"))
            self.ui.check_chamfer.show()

            # TODO: change to Qt5 style
            QtCore.QObject.connect(self.ui.check_delete,
                                   QtCore.SIGNAL("stateChanged(int)"),
                                   self.set_delete)
            QtCore.QObject.connect(self.ui.check_chamfer,
                                   QtCore.SIGNAL("stateChanged(int)"),
                                   self.set_chamfer)

            # TODO: somehow we need to set up the trackers
            # to show a preview of the fillet.

            # self.linetrack = trackers.lineTracker(dotted=True)
            # self.arctrack = trackers.arcTracker()
            # self.call = self.view.addEventCallback("SoEvent", self.action)
            _msg(translate("draft","Enter radius."))

    def action(self, arg):
        """Scene event handler. CURRENTLY NOT USED.

        Here the displaying of the trackers (previews)
        should be implemented by considering the current value of the
        `ui.radiusValue`.
        """
        if arg["Type"] == "SoKeyboardEvent":
            if arg["Key"] == "ESCAPE":
                self.finish()
        elif arg["Type"] == "SoLocation2Event":
            self.point, ctrlPoint, info = gui_tool_utils.getPoint(self, arg)
            gui_tool_utils.redraw3DView()

    def set_delete(self):
        """Execute as a callback when the delete checkbox changes."""
        self.delete = self.ui.check_delete.isChecked()
        _msg(translate("draft","Delete original objects:") + " " + str(self.delete))

    def set_chamfer(self):
        """Execute as a callback when the chamfer checkbox changes."""
        self.chamfer = self.ui.check_chamfer.isChecked()
        _msg(translate("draft","Chamfer mode:") + " " + str(self.chamfer))

    def numericRadius(self, rad):
        """Validate the entry radius in the user interface.

        This function is called by the toolbar or taskpanel interface
        when a valid radius has been entered in the input field.
        """
        self.rad = rad
        self.draw_arc(rad, self.chamfer, self.delete)
        self.finish()

    def draw_arc(self, rad, chamfer, delete):
        """Process the selection and draw the actual object."""
        wires = Gui.Selection.getSelection()

        if not wires or len(wires) != 2:
            _err(translate("draft","Two elements needed."))
            return

        for o in wires:
            _msg(utils.get_type(o))

        _test = translate("draft", "Test object")
        _test_off = translate("draft", "Test object removed")
        _cant = translate("draft", "Fillet cannot be created")

        _msg(4*"=" + _test)
        arc = Draft.make_fillet(wires, rad)
        if not arc:
            _err(_cant)
            return
        self.doc.removeObject(arc.Name)
        _msg(4*"=" + _test_off)

        _doc = 'FreeCAD.ActiveDocument.'

        _wires = '['
        _wires += _doc + wires[0].Name + ', '
        _wires += _doc + wires[1].Name
        _wires += ']'

        Gui.addModule("Draft")

        _cmd = 'Draft.make_fillet'
        _cmd += '('
        _cmd += _wires + ', '
        _cmd += 'radius=' + str(rad)
        if chamfer:
            _cmd += ', chamfer=' + str(chamfer)
        if delete:
            _cmd += ', delete=' + str(delete)
        _cmd += ')'
        _cmd_list = ['arc = ' + _cmd,
                     'Draft.autogroup(arc)',
                     'FreeCAD.ActiveDocument.recompute()']

        self.commit(translate("draft", "Create fillet"),
                    _cmd_list)

    def finish(self, close=False):
        """Terminate the operation."""
        super(Fillet, self).finish()
        if self.ui:
            # self.linetrack.finalize()
            # self.arctrack.finalize()
            self.doc.recompute()


Gui.addCommand('Draft_Fillet', Fillet())

## @}



# ***************************************************************************
# *   Copyright (c) 2014 Yorik van Havre <yorik@uncreated.net>              *
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the object code for the Layer object."""
## @package layer
# \ingroup draftobjects
# \brief Provides the object code for the Layer object.

## \addtogroup draftobjects
# @{
from PySide.QtCore import QT_TRANSLATE_NOOP


class Layer:
    """The Layer object.

    This class is normally used to extend a base `App::FeaturePython` object.
    """

    def __init__(self, obj):
        self.Type = "Layer"
        self.Object = obj
        self.set_properties(obj)

        obj.Proxy = self

    def onDocumentRestored(self, obj):
        """Execute code when the document is restored.

        Add properties that don't exist.
        """
        self.set_properties(obj)

    def set_properties(self, obj):
        """Set properties only if they don't exist."""
        if "Group" not in obj.PropertiesList:
            _tip = QT_TRANSLATE_NOOP("App::Property",
                                     "The objects that are part of this layer")
            obj.addProperty("App::PropertyLinkList",
                            "Group",
                            "Layer",
                            _tip)

    def __getstate__(self):
        """Return a tuple of objects to save or None."""
        return self.Type

    def __setstate__(self, state):
        """Set the internal properties from the restored state."""
        if state:
            self.Type = state

    def execute(self, obj):
        """Execute when the object is created or recomputed. Do nothing."""
        pass

    def addObject(self, obj, child):
        """Add an object to this object if not in the Group property."""
        group = obj.Group
        if child not in group:
            group.append(child)
        obj.Group = group


# Alias for compatibility with v0.18 and earlier
_VisGroup = Layer


class LayerContainer:
    """The container object for layers.

    This class is normally used to extend
    a base `App::DocumentObjectGroupPython` object.
    """

    def __init__(self, obj):
        self.Type = "LayerContainer"
        obj.Proxy = self

    def execute(self, obj):
        """Execute when the object is created or recomputed.

        Update the value of `Group` by sorting the contained layers
        by `Label`.
        """
        group = obj.Group
        group.sort(key=lambda layer: layer.Label)
        obj.Group = group

    def __getstate__(self):
        """Return a tuple of objects to save or None."""
        if hasattr(self, "Type"):
            return self.Type

    def __setstate__(self, state):
        """Set the internal properties from the restored state."""
        if state:
            self.Type = state

## @}
